from collections import defaultdict
from typing import Mapping, Union

from sqlalchemy import select

from vigorish.config.database import (
    Pitch_Mix_By_Year_View,
    Pitch_Type_All_View,
    Pitch_Type_By_Year_View,
    Player,
    Season,
)
from vigorish.enums import PitchType
from vigorish.util.pitch_calcs import calc_pitch_mix, get_pitch_type_metrics

PitchMix = Mapping[PitchType, Mapping[str, Union[int, float]]]


class AllPlayerData:
    def __init__(self, app, mlb_id):
        self.app = app
        self.db_engine = self.app.db_engine
        self.db_session = self.app.db_session
        self.mlb_id = mlb_id
        self.player = Player.find_by_mlb_id(self.db_session, self.mlb_id)

    @property
    def years_played(self):
        s = select([Pitch_Mix_By_Year_View]).where(Pitch_Mix_By_Year_View.id == self.player.id)
        results = self.db_engine.execute(s).fetchall()
        years_played = []
        for d in [dict(row) for row in results]:
            season = self.db_session.query(Season).get(d["season_id"])
            years_played.append(season.year)
        return sorted(years_played)

    @property
    def pitch_mix(self):
        return self.player.pitch_mix

    @property
    def pitch_mix_right(self):
        return self.player.pitch_mix_right

    @property
    def pitch_mix_left(self):
        return self.player.pitch_mix_left

    @property
    def pitch_mix_by_year(self):
        s = select([Pitch_Mix_By_Year_View]).where(Pitch_Mix_By_Year_View.id == self.player.id)
        results = self.db_engine.execute(s).fetchall()
        pitch_mix = {}
        for d in [dict(row) for row in results]:
            season = self.db_session.query(Season).get(d["season_id"])
            pitch_mix[season.year] = calc_pitch_mix(d)
        return pitch_mix

    @property
    def pitch_appearances(self):
        return self.player.pitch_apps

    @property
    def pitch_types(self):
        s = select([Pitch_Type_All_View]).where(Pitch_Type_All_View.id == self.player.id)
        results = self.db_engine.execute(s).fetchall()
        pitch_types = {}
        for d in [dict(row) for row in results]:
            (pitch_type, metrics) = get_pitch_type_metrics(d)
            pitch_types[pitch_type] = metrics
        return pitch_types

    @property
    def pitch_types_by_year(self):
        s = select([Pitch_Type_By_Year_View]).where(Pitch_Type_By_Year_View.id == self.player.id)
        results = self.db_engine.execute(s).fetchall()
        pitch_types = defaultdict(dict)
        for d in [dict(row) for row in results]:
            season = self.db_session.query(Season).get(d["season_id"])
            (pitch_type, metrics) = get_pitch_type_metrics(d)
            pitch_types[season.year][pitch_type] = metrics
        return pitch_types



from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session, sessionmaker

from vigorish.config.config_file import ConfigFile
from vigorish.config.database import delete_sqlite_database, get_db_url
from vigorish.config.dotenv_file import DotEnvFile
from vigorish.data.scraped_data import ScrapedData
from vigorish.util.result import Result


class Vigorish:
    dotenv: DotEnvFile
    config: ConfigFile
    db_engine: Engine
    db_session: Session
    scraped_data: ScrapedData

    def __init__(self, db_url=None) -> None:
        self.initialize_app(db_url)

    def initialize_app(self, db_url):
        self.dotenv = DotEnvFile()
        self.config = ConfigFile()
        if not db_url:
            db_url = get_db_url()
        self.db_engine = create_engine(db_url)
        session_maker = sessionmaker(bind=self.db_engine)
        self.db_session = session_maker()
        self.scraped_data = ScrapedData(self.db_engine, self.db_session, self.config)

    def reset_database_connection(self, db_url=None):
        if not db_url:
            db_url = get_db_url()
        result = delete_sqlite_database(db_url)
        if result.failure:
            return result
        self.db_session.close()
        self.db_session = None
        self.initialize_app(db_url)
        return Result.Ok()



"""Custom exceptions for issues specific to vigorish actions/processes."""


from vigorish.enums import DataSet


class ConfigSetingException(Exception):
    """Exception raised when an unrecoverable error occurs due to a ConfigSetting."""

    def __init__(self, setting_name, data_set=DataSet.ALL, current_value=None, detail=None):
        message = (
            f"ConfigSetting value is invalid:\n"
            f"\tName...........: {setting_name} (Data Set: {data_set})\n"
        )
        if current_value:
            message += f"\tCurrent Value..: {current_value}\n"
        if detail:
            message += f"\tDetail.........: {detail}\n"
        super().__init__(message)


class ScrapedDataException(Exception):
    """Exception raised when data identified by file_type, data_set and url_id cannot be found."""

    def __init__(self, file_type, data_set, url_id):
        message = (
            f"Failed to locate scraped data: URL ID: {url_id} "
            f"(File Type, {file_type}, Data Set: {data_set})"
        )
        super().__init__(message)



from copy import deepcopy

from vigorish.enums import PitchType


def calc_pitch_mix(pitch_count_dict, threshold=0.01):
    pitch_types = []
    mix_total = 0
    for col_name, val in pitch_count_dict.items():
        if not ("_count" in col_name):
            continue
        if val == 0:
            continue
        if not _check_threshold(val, pitch_count_dict["total_pitches"], threshold):
            continue
        mix_total += val
        pitch_type_abbrev = col_name.split("_")[0].upper()
        pitch_type = PitchType.from_abbrev(pitch_type_abbrev)
        pitch_types.append({"type": pitch_type, "count": val})
    pitch_mix_int = 0
    for pt in pitch_types:
        pt["percent"] = round(pt["count"] / float(mix_total), 3)
        pitch_mix_int += int(pt["type"])
    pitch_types.append({"type": PitchType.ALL, "percent": 1.000, "count": mix_total})
    pitch_types.sort(key=lambda x: x["percent"], reverse=True)
    return {x["type"]: {"percent": x["percent"], "count": x["count"]} for x in pitch_types}


def _check_threshold(pitch_count, total_pitches, threshold):
    percent = pitch_count / float(total_pitches)
    return percent >= threshold


def get_pitch_mix_from_int(pitch_mix_int):
    pitch_types = []
    for ptype in PitchType:
        if ptype == PitchType.ALL:
            continue
        if pitch_mix_int & ptype == ptype:
            pitch_types.append(ptype)
    return pitch_types


def get_pitch_type_metrics(pitch_type_dict):
    d = deepcopy(pitch_type_dict)
    d.pop("id")
    d.pop("season_id")
    d.pop("mlb_id")
    d.pop("name")
    pitch_type = PitchType.from_abbrev(d.pop("pitch_type"))
    for metric, val in d.items():
        d[metric] = round(val, 3)
    return (pitch_type, d)
